from unittest.mock import patch

import pytest
from orchestrator.types import SubscriptionLifecycle

from gso.products import ProductName
from gso.products.product_types.lan_switch_interconnect import LanSwitchInterconnect
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.types.ip_address import AddressSpace
from test.services.conftest import MockedNetboxClient
from test.workflows import assert_complete, extract_state, run_workflow


@pytest.fixture()
def _netbox_client_mock():
    # Mock NetboxClient methods
    with (
        patch("gso.services.netbox_client.NetboxClient.get_device_by_name") as mock_get_device_by_name,
        patch("gso.services.netbox_client.NetboxClient.get_available_interfaces") as mock_get_available_interfaces,
        patch("gso.services.netbox_client.NetboxClient.get_available_lags_in_range") as mock_available_lags_in_range,
    ):
        mock_get_device_by_name.return_value = MockedNetboxClient().get_device_by_name()
        mock_get_available_interfaces.return_value = MockedNetboxClient().get_available_interfaces()
        mock_available_lags_in_range.return_value = MockedNetboxClient().get_available_lags_in_range()

        yield


@pytest.fixture()
def input_form_data(faker, router_subscription_factory, switch_subscription_factory):
    def _generate_form_data(address_space: AddressSpace):
        return [
            {
                "product": get_product_id_by_name(ProductName.LAN_SWITCH_INTERCONNECT),
            },
            {
                "tt_number": faker.tt_number(),
                "router_side": router_subscription_factory(),
                "switch_side": switch_subscription_factory(),
                "address_space": address_space,
                "description": faker.sentence(),
                "minimum_link_count": 2,
                "vlan_id": 111,  # VLAN ID for new interconnections is always 111
            },
            {
                "router_side_iface": "lag-1",
                "router_side_ae_members": faker.link_members_nokia()[:2],
            },
            {
                "switch_side_iface": faker.network_interface(),
                "switch_side_ae_members": faker.link_members_juniper()[:2],
            },
            {},
        ]

    return _generate_form_data


@pytest.mark.parametrize("address_space", [AddressSpace.PRIVATE, AddressSpace.PUBLIC])
@pytest.mark.workflow()
def test_create_lan_switch_interconnect_success(
    address_space,
    input_form_data,
    _netbox_client_mock,  # noqa: PT019
):
    initial_data = input_form_data(address_space)

    result, _, _ = run_workflow("create_lan_switch_interconnect", initial_data)

    assert_complete(result)
    state = extract_state(result)
    subscription_id = state["subscription_id"]
    subscription = LanSwitchInterconnect.from_subscription(subscription_id)
    assert subscription.status == SubscriptionLifecycle.ACTIVE
